@@ -9,6 +9,7 @@ import logging
 import os
 import shlex
 import shutil
+import struct
 import random
 import time
 import types
@@ -206,6 +207,8 @@ def environment(ctx, build, vary):
             removed += [k]
         else:
             added += [(k, v)]
+    if ctx.verbosity >= 1:
+        logger.info("ENVIRONMENT variation: added = (" + ','.join(map(str, added)) + ") removed = " + ''.join(removed))
     return build.modify_env(added, removed)
 
 def domain_host(ctx, build, vary):
@@ -214,6 +217,8 @@ def domain_host(ctx, build, vary):
     hostname = "reprotest-capture-hostname"
     domainname = "reprotest-capture-domainname"
     _ = build
+    if ctx.verbosity >= 1:
+        logger.info("DOMAIN_HOST variation: hostname = " + hostname + " domainname = " + domainname)
 
     # TODO: below only works on linux, of course..
     if ctx.spec.domain_host.use_sudo:
@@ -249,12 +254,16 @@ def build_path(ctx, build, vary):
     if vary:
         return build
     const_path = os.path.join(dirname(build.tree), 'const_build_path')
+    if ctx.verbosity >= 1:
+        logger.info("BUILD_PATH variation: const_path = " + const_path)
     return build.move_tree(build.tree, const_path, True)
 
 @tool_required("disorderfs")
 def fileordering(ctx, build, vary):
     if not vary:
         return build
+    if ctx.verbosity >= 1:
+        logger.info("DISORDERFS variation: --shuffle-dirents=yes")
 
     old_tree = os.path.join(dirname(build.tree), basename(build.tree) + '-before-disorderfs', '')
     _ = build.move_tree(build.tree, old_tree, False)
@@ -274,6 +283,8 @@ def home(ctx, build, vary):
         # choose an existent HOME, see Debian bug #860428
         return build.add_env('HOME', build.tree)
     else:
+        if ctx.verbosity >= 1:
+            logger.info("HOME variation: HOME = /nonexistent/second-build")
         return build.add_env('HOME', '/nonexistent/second-build')
 
 # TODO: uname is a POSIX standard.  The related Linux command
@@ -286,13 +297,37 @@ def kernel(ctx, build, vary):
     if not vary:
         _ = _.append_setup_exec_raw('SETARCH_ARCH=$(uname -m)')
     else:
-        _ = _.append_setup_exec_raw('SETARCH_ARCH=$(setarch --list | grep -vF "$(uname -m)" | shuf | head -n1)')
+        _ = _.append_setup_exec_raw('SETARCH_ARCH=$(for a in $(setarch --list); do setarch $a true && echo $a || true; done)')
+        # Perform realistic shuffling of architectures depending
+        # on the bitness of application.
+        # 64-bit binaries can NOT be executed on 32-bit CPU,
+        # so filter out setarch's PER_LINUX32 architectures:
+        #
+        # echo $(grep -ioe "{PER_LINUX.*,.*$" /tmp/setarch.c |
+        #     grep -v machine | grep PER_LINUX32 | awk '{print $2}' | sort -Vu) | sed 's/,//g;s/"//g'
+        _ = _.append_setup_exec_raw('DROP_ARCH="-v -e ^$(uname -m)\$"')
+        _ = _.append_setup_exec_raw('WORDSIZE={}'.format(struct.calcsize("P")*8))
+        _ = _.append_setup_exec_raw('if [ $WORDSIZE -eq 64 ]; then \
+    for _ARCH_TO_DROP in armh armv7b armv7l \
+        armv8b armv8l arm athlon i386 i486 \
+        i586 i686 linux32 mips32 mips parisc32 \
+        parisc ppc32le ppc32 ppcle ppc s390 \
+        sparc32bash sparc32 sparc; \
+    do \
+        DROP_ARCH="$DROP_ARCH -e ^$_ARCH_TO_DROP\$"; \
+    done; \
+fi')
+        _ = _.append_setup_exec_raw('SETARCH_ARCH=$(echo "$SETARCH_ARCH" | grep $DROP_ARCH | shuf -n1)')
         _ = _.append_setup_exec_raw('KERNEL_VERSION=$(uname -r)')
         _ = _.append_setup_exec_raw('if [ ${KERNEL_VERSION#2.6} = $KERNEL_VERSION ]; then SETARCH_OPTS=--uname-2.6; fi')
+        if ctx.verbosity >= 1:
+            _ = _.append_setup_exec_raw('echo "INFO:reprotest.build:KERNEL variation: SETARCH_ARCH = \'$SETARCH_ARCH\' SETARCH_OPTS = \'$SETARCH_OPTS\'" >&2')
     return _.prepend_to_build_command_raw('setarch', '$SETARCH_ARCH', '$SETARCH_OPTS')
 
 def aslr(ctx, build, vary):
     if vary:
+        if ctx.verbosity >= 1:
+            logger.info("ASLR variation: ASLR enabled")
         return build
     return build.append_setup_exec_raw('SETARCH_OPTS="$SETARCH_OPTS -R"')
 
@@ -311,8 +346,10 @@ def num_cpus(ctx, build, vary):
             else shuf -i$((CPU_MIN + 1))-$CPU_MAX -n1; fi)')
 
     # select CPU_NUM random cpus from the range 0..$((CPU_MAX-1))
-    cpu_list = "$(echo $(shuf -i0-$((CPU_MAX - 1)) -n$CPU_NUM) | tr ' ' ,)"
-    return _.prepend_to_build_command_raw('taskset', '-a', '-c', cpu_list)
+    _ = _.append_setup_exec_raw("export CPU_LIST=\"$(echo $(shuf -i0-$((CPU_MAX - 1)) -n$CPU_NUM) | tr ' ' ,)\"")
+    if vary and ctx.verbosity >= 1:
+        _ = _.append_setup_exec_raw('echo "INFO:reprotest.build:NUM_CPUS variation: cpu_list = $CPU_LIST" >&2')
+    return _.prepend_to_build_command_raw('taskset', '-a', '-c', '$CPU_LIST')
 
 # TODO: if this locale doesn't exist on the system, Python's
 # locales.getlocale() will return (None, None) rather than this
@@ -330,14 +367,19 @@ def locales(ctx, build, vary):
     if not vary:
         return build.add_env('LANG', 'C.UTF-8').add_env('LANGUAGE', 'en_US:en')
     else:
-        # if there is an issue with this being random, we could instead select it
-        # based on a deterministic hash of the inputs
-        loc = random.choice(['fr_CH.UTF-8', 'es_ES', 'ru_RU.CP1251', 'kk_KZ.RK1048', 'zh_CN'])
+        # Use an Estonian locale by default, which has interesting sort order for ASCII characters
+        loc = 'et_EE.UTF-8'
+        if ctx.spec.locales.locale:
+            loc = random.choice(ctx.spec.locales.locale)
+        if ctx.verbosity >= 1:
+            logger.info("LOCALE variation: LC_ALL = " + loc + " LANGUAGE = " + loc + ":fr")
         return build.add_env('LANG', loc).add_env('LC_ALL', loc).add_env('LANGUAGE', '%s:fr' % loc)
 
 def exec_path(ctx, build, vary):
     if not vary:
         return build
+    if ctx.verbosity >= 1:
+        logger.info("EXEC_PATH variation: PATH = " + build.env['PATH'] + ':/i_capture_the_path')
     return build.add_env('PATH', build.env['PATH'] + ':/i_capture_the_path')
 
 # This doesn't require superuser privileges, but the chsh command
@@ -351,8 +393,12 @@ def timezone(ctx, build, vary):
     # (http://pubs.opengroup.org/onlinepubs/9699919799/basedefs/V1_chap08.html#tag_08),
     # so they should be cross-platform compatible.
     if not vary:
+        if ctx.verbosity >= 1:
+            logger.info("TIMEZONE variation: TZ=GMT+12")
         return build.add_env('TZ', 'GMT+12')
     else:
+        if ctx.verbosity >= 1:
+            logger.info("TIMEZONE variation: TZ=GMT-14")
         return build.add_env('TZ', 'GMT-14')
 
 @tool_required("faketime")
@@ -371,6 +417,8 @@ def faketime(ctx, build, vary):
         if base < now - 15552000 and not random.randint(0, 1):
             # if ctx.base_faketime is far in the past, with 1/2 probability
             # reuse the current time and don't fake it
+            if ctx.verbosity >= 1:
+                logger.info("FAKETIME variation: enabled but chosen randomly not to fake!")
             return build
         else:
             # otherwise use a date far in the future
@@ -381,12 +429,16 @@ def faketime(ctx, build, vary):
     # this is only mentioned in the README. we do not want this, it really really
     # messes with GNU make and other buildsystems that look at timestamps.
     # set NO_FAKE_STAT=1 avoids this.
+    if ctx.verbosity >= 1:
+        logger.info("FAKETIME variation: faketime = " + faket)
     return build.add_env('NO_FAKE_STAT', '1').prepend_to_build_command('faketime', faket)
 
 def umask(ctx, build, vary):
     if not vary:
         return build.append_setup_exec('umask', '0022')
     else:
+        if ctx.verbosity >= 1:
+            logger.info("UMASK variation: umask 0002")
         return build.append_setup_exec('umask', '0002')
 
 
@@ -525,6 +577,10 @@ class DomainHostVariation(collections.namedtuple('_DomainHostVariation', 'use_su
     def default(cls):
         return cls(0)
 
+class LocalesVariation(collections.namedtuple('_LocaleVariation', 'locale')):
+    @classmethod
+    def default(cls):
+        return cls(mdiffconf.strlist_set(";"))
 
 class VariationSpec(mdiffconf.ImmutableNamespace):
     @classmethod
@@ -534,6 +590,7 @@ class VariationSpec(mdiffconf.ImmutableNamespace):
             "user_group": UserGroupVariation.default(),
             "time": TimeVariation.default(),
             "domain_host": DomainHostVariation.default(),
+            "locales": LocalesVariation.default(),
         }
         return cls(**{k: default_overrides.get(k, True) for k in variations})
 
